# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

from .... import tests
from ..helpers import kind_to_mode
from . import FastimportFeature

try:
    from fastimport import commands
except ImportError:
    commands = object()


def load_tests(loader, standard_tests, pattern):
    """Parameterize tests for all versions of groupcompress."""
    scenarios = [
        ('pack-0.92', {'branch_format': 'pack-0.92'}),
        ('1.9-rich-root', {'branch_format': '1.9-rich-root'}),
    ]
    from ....bzr.groupcompress_repo import RepositoryFormat2a
    scenarios.append(('2a', {'branch_format': '2a'}))
    suite = loader.suiteClass()
    result = tests.multiply_tests(standard_tests, scenarios, suite)
    return result


class TestCaseForGenericProcessor(tests.TestCaseWithTransport):

    _test_needs_features = [FastimportFeature]

    branch_format = "pack-0.92"

    def get_handler(self):
        from ..processors import generic_processor
        branch = self.make_branch('.', format=self.branch_format)
        handler = generic_processor.GenericProcessor(branch.controldir)
        return handler, branch

    # FIXME: [] as a default is bad, as it is mutable, but I want
    # to use None to mean "don't check this".
    def assertChanges(self, branch, revno, expected_added=[],
                      expected_removed=[], expected_modified=[],
                      expected_renamed=[], expected_kind_changed=[]):
        """Check the changes introduced in a revision of a branch.

        This method checks that a revision introduces expected changes.
        The required changes are passed in as a list, where
        each entry contains the needed information about the change.

        If you do not wish to assert anything about a particular
        category then pass None instead.

        branch: The branch.
        revno: revision number of revision to check.
        expected_added: a list of (filename,) tuples that must have
            been added in the delta.
        expected_removed: a list of (filename,) tuples that must have
            been removed in the delta.
        expected_modified: a list of (filename,) tuples that must have
            been modified in the delta.
        expected_renamed: a list of (old_path, new_path) tuples that
            must have been renamed in the delta.
        expected_kind_changed: a list of (path, old_kind, new_kind) tuples
            that must have been changed in the delta.
        :return: revtree1, revtree2
        """
        repo = branch.repository
        revtree1 = repo.revision_tree(branch.get_rev_id(revno - 1))
        revtree2 = repo.revision_tree(branch.get_rev_id(revno))
        changes = revtree2.changes_from(revtree1)
        self._check_changes(changes, expected_added, expected_removed,
                            expected_modified, expected_renamed, expected_kind_changed)
        return revtree1, revtree2

    def _check_changes(self, changes, expected_added=[],
                       expected_removed=[], expected_modified=[],
                       expected_renamed=[], expected_kind_changed=[]):
        """Check the changes in a TreeDelta

        This method checks that the TreeDelta contains the expected
        modifications between the two trees that were used to generate
        it. The required changes are passed in as a list, where
        each entry contains the needed information about the change.

        If you do not wish to assert anything about a particular
        category then pass None instead.

        changes: The TreeDelta to check.
        expected_added: a list of (filename,) tuples that must have
            been added in the delta.
        expected_removed: a list of (filename,) tuples that must have
            been removed in the delta.
        expected_modified: a list of (filename,) tuples that must have
            been modified in the delta.
        expected_renamed: a list of (old_path, new_path) tuples that
            must have been renamed in the delta.
        expected_kind_changed: a list of (path, old_kind, new_kind) tuples
            that must have been changed in the delta.
        """
        renamed = changes.renamed
        added = changes.added + changes.copied
        removed = changes.removed
        modified = changes.modified
        kind_changed = changes.kind_changed
        if expected_renamed is not None:
            self.assertEqual(len(renamed), len(expected_renamed),
                              "{} is renamed, expected {}".format(renamed, expected_renamed))
            renamed_files = [(item.path[0], item.path[1]) for item in renamed]
            for expected_renamed_entry in expected_renamed:
                expected_renamed_entry = (
                    expected_renamed_entry[0].decode('utf-8'),
                    expected_renamed_entry[1].decode('utf-8'))
                self.assertTrue(expected_renamed_entry in renamed_files,
                                "{} is not renamed, {} are".format(expected_renamed_entry,
                                                               renamed_files))
        if expected_added is not None:
            self.assertEqual(len(added), len(expected_added),
                              "%s is added" % str(added))
            added_files = [(item.path[1],) for item in added]
            for expected_added_entry in expected_added:
                expected_added_entry = (
                    expected_added_entry[0].decode('utf-8'), )
                self.assertTrue(expected_added_entry in added_files,
                                "{} is not added, {} are".format(expected_added_entry,
                                                             added_files))
        if expected_removed is not None:
            self.assertEqual(len(removed), len(expected_removed),
                              "%s is removed" % str(removed))
            removed_files = [(item.path[0],) for item in removed]
            for expected_removed_entry in expected_removed:
                expected_removed_entry = (
                    expected_removed_entry[0].decode('utf-8'), )
                self.assertTrue(expected_removed_entry in removed_files,
                                "{} is not removed, {} are".format(expected_removed_entry,
                                                               removed_files))
        if expected_modified is not None:
            self.assertEqual(len(modified), len(expected_modified),
                              "%s is modified" % str(modified))
            modified_files = [(item.path[1],) for item in modified]
            for expected_modified_entry in expected_modified:
                expected_modified_entry = (
                    expected_modified_entry[0].decode('utf-8'), )
                self.assertTrue(expected_modified_entry in modified_files,
                                "{} is not modified, {} are".format(
                                    expected_modified_entry, modified_files))
        if expected_kind_changed is not None:
            self.assertEqual(len(kind_changed), len(expected_kind_changed),
                              "{} is kind-changed, expected {}".format(kind_changed,
                                                                   expected_kind_changed))
            kind_changed_files = [(item.path[1], item.kind[0], item.kind[1])
                                  for item in kind_changed]
            for expected_kind_changed_entry in expected_kind_changed:
                expected_kind_changed_entry = (
                    expected_kind_changed_entry[0].decode('utf-8'), ) + expected_kind_changed_entry[1:]
                self.assertTrue(expected_kind_changed_entry in
                                kind_changed_files, "{} is not kind-changed, {} are".format(
                                    expected_kind_changed_entry, kind_changed_files))

    def assertContent(self, branch, tree, path, content):
        with branch.lock_read():
            self.assertEqual(tree.get_file_text(path.decode('utf-8')), content)

    def assertSymlinkTarget(self, branch, tree, path, target):
        with branch.lock_read():
            self.assertEqual(tree.get_symlink_target(
                path.decode('utf-8')), target)

    def assertExecutable(self, branch, tree, path, executable):
        with branch.lock_read():
            self.assertEqual(tree.is_executable(
                path.decode('utf-8')), executable)

    def assertRevisionRoot(self, revtree, path):
        self.assertEqual(revtree.get_revision_id(),
                         revtree.get_file_revision(path.decode('utf-8')))


class TestImportToPackTag(TestCaseForGenericProcessor):

    def file_command_iter(self, path, kind='file', content=b'aaa',
                          executable=False, to_kind=None, to_content=b'bbb', to_executable=None):
        # Revno 1: create a file or symlink
        # Revno 2: modify it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path,
                                                 kind_to_mode(kind, executable), None, content)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(path,
                                                 kind_to_mode(to_kind, to_executable), None, to_content)

            # pass "head" for from_ to show that #401249 is worse than I knew
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b"head", [], files_two)

            yield commands.TagCommand(b'tag1', b':1', committer, b"tag 1")

            # pass "head" for from_ to demonstrate #401249
            yield commands.TagCommand(b'tag2', b'head', committer, b"tag 2")
        return command_list

    def test_tag(self):
        handler, branch = self.get_handler()
        path = b'a'
        raise tests.KnownFailure("non-mark committish not yet supported"
                                 "- bug #410249")
        handler.process(self.file_command_iter(path))


class TestImportZeroMarker(TestCaseForGenericProcessor):

    def test_tag(self):
        handler, branch = self.get_handler()

        def command_list():
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]
            yield commands.TagCommand(b'tag1', b':0', committer, b"tag 1")
        handler.process(command_list)


class TestImportToPackModify(TestCaseForGenericProcessor):

    def file_command_iter(self, path, kind='file', content=b'aaa',
                          executable=False, to_kind=None, to_content=b'bbb', to_executable=None):

        # Revno 1: create a file or symlink
        # Revno 2: modify it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable
        mode = kind_to_mode(kind, executable)
        to_mode = kind_to_mode(to_kind, to_executable)

        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, mode, None, content)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(path, to_mode, None, to_content)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_modify_file_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")

    def test_modify_file_becomes_symlink(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path,
                                               kind='file', to_kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_kind_changed=[(path, 'file', 'symlink')])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")

    def test_modify_symlink_becomes_file(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path,
                                               kind='symlink', to_kind='file'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_kind_changed=[(path, 'symlink', 'file')])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_modify_file_now_executable(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path,
                                               executable=False, to_executable=True, to_content=b'aaa'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertExecutable(branch, revtree1, path, False)
        self.assertExecutable(branch, revtree2, path, True)

    def test_modify_file_no_longer_executable(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path,
                                               executable=True, to_executable=False, to_content=b'aaa'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[(path,)])
        self.assertExecutable(branch, revtree1, path, True)
        self.assertExecutable(branch, revtree2, path, False)


class TestImportToPackModifyTwice(TestCaseForGenericProcessor):
    """This tests when the same file is modified twice in the one commit.

    Note: hg-fast-export produces data like this on occasions.
    """

    def file_command_iter(self, path, kind='file', content=b'aaa',
                          executable=False, to_kind=None, to_content=b'bbb', to_executable=None):

        # Revno 1: create a file twice
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, kind_to_mode(kind, executable),
                                                 None, content)
                yield commands.FileModifyCommand(path, kind_to_mode(to_kind, to_executable),
                                                 None, to_content)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)
        return command_list

    def test_modify_file_twice_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertRevisionRoot(revtree1, path)


class TestImportToPackModifyTricky(TestCaseForGenericProcessor):

    def file_command_iter(self, path1, path2, kind='file'):

        # Revno 1: create a file or symlink in a directory
        # Revno 2: create a second file that implicitly deletes the
        # first one because either:
        # * the new file is a in directory with the old file name
        # * the new file has the same name as the directory of the first
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path1, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(path2, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_modify_file_becomes_directory(self):
        handler, branch = self.get_handler()
        path1 = b'a/b'
        path2 = b'a/b/c'
        handler.process(self.file_command_iter(path1, path2))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path1,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(path2,)],
                                                expected_kind_changed=[(path1, 'file', 'directory')])
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, path2, b"bbb")

    def test_modify_directory_becomes_file(self):
        handler, branch = self.get_handler()
        path1 = b'a/b/c'
        path2 = b'a/b'
        handler.process(self.file_command_iter(path1, path2))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (path1,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path1,), ],
                                                expected_kind_changed=[(path2, 'directory', 'file')])
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, path2, b"bbb")

    def test_modify_symlink_becomes_directory(self):
        handler, branch = self.get_handler()
        path1 = b'a/b'
        path2 = b'a/b/c'
        handler.process(self.file_command_iter(path1, path2, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path1,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(path2,)],
                                                expected_kind_changed=[(path1, 'symlink', 'directory')])
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path2, "bbb")

    def test_modify_directory_becomes_symlink(self):
        handler, branch = self.get_handler()
        path1 = b'a/b/c'
        path2 = b'a/b'
        handler.process(self.file_command_iter(path1, path2, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (path1,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path1,), ],
                                                expected_kind_changed=[(path2, 'directory', 'symlink')])
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path2, "bbb")


class TestImportToPackDelete(TestCaseForGenericProcessor):

    def file_command_iter(self, path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: delete it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_file_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertRevisionRoot(revtree1, path)

    def test_delete_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(b'a',), (path,)])
        self.assertContent(branch, revtree1, path, b"aaa")

    def test_delete_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertRevisionRoot(revtree1, path)

    def test_delete_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(b'a',), (path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")

    def test_delete_file_in_deep_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/b/c/d'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (b'a/b/c',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(b'a',), (b'a/b',), (b'a/b/c',), (path,)])
        self.assertContent(branch, revtree1, path, b"aaa")


class TestImportToPackDeleteNew(TestCaseForGenericProcessor):
    """Test deletion of a newly added file."""

    def file_command_iter(self, path, kind='file'):

        # Revno 1: create a file or symlink then delete it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileDeleteCommand(path)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)
        return command_list

    def test_delete_new_file_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,)

    def test_delete_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,)

    def test_delete_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,)

    def test_delete_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,)

    def test_delete_new_file_in_deep_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/b/c/d'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,)


class TestImportToPackDeleteMultiLevel(TestCaseForGenericProcessor):

    def file_command_iter(self, paths, paths_to_delete):

        # Revno 1: create multiple files
        # Revno 2: delete multiple files
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(path, kind_to_mode('file', False),
                                                     None, b"aaa%d" % i)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                for path in paths_to_delete:
                    yield commands.FileDeleteCommand(path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_files_in_multiple_levels(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d/e']
        paths_to_delete = [b'a/b/c', b'a/b/d/e']
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[
                                                    (b'a',), (b'a/b',
                                                              ), (b'a/b/c',),
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a',), (b'a/b',
                                                              ), (b'a/b/c',),
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    ])

    def test_delete_file_single_level(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d/e']
        paths_to_delete = [b'a/b/d/e']
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[
                                                    (b'a',), (b'a/b',
                                                              ), (b'a/b/c',),
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    ])

    def test_delete_file_complex_level(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d/e', b'a/f/g', b'a/h', b'a/b/d/i/j']
        paths_to_delete = [b'a/b/c', b'a/b/d/e', b'a/f/g', b'a/b/d/i/j']
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[
                                                    (b'a',), (b'a/b',
                                                              ), (b'a/b/c',),
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    (b'a/f',), (b'a/f/g',),
                                                    (b'a/h',),
                                                    (b'a/b/d/i',), (b'a/b/d/i/j',),
                                                    ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a/b',), (b'a/b/c',),
                                                    (b'a/b/d',), (b'a/b/d/e',),
                                                    (b'a/f',), (b'a/f/g',),
                                                    (b'a/b/d/i',), (b'a/b/d/i/j',),
                                                    ])


class TestImportToPackDeleteThenAdd(TestCaseForGenericProcessor):
    """Test delete followed by an add. Merges can cause this."""

    def file_command_iter(self, path, kind='file', content=b'aaa',
                          executable=False, to_kind=None, to_content=b'bbb', to_executable=None):

        # Revno 1: create a file or symlink
        # Revno 2: delete it and add it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, kind_to_mode(kind, executable),
                                                 None, content)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(path)
                yield commands.FileModifyCommand(path, kind_to_mode(to_kind, to_executable),
                                                 None, to_content)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_then_add_file_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)],
                                                expected_added=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_delete_then_add_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)],
                                                expected_added=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_delete_then_add_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b'a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)],
                                                expected_added=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_delete_then_add_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b'a/a'
        handler.process(self.file_command_iter(path, kind='symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(path,)],
                                                expected_added=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")


class TestImportToPackDeleteDirectory(TestCaseForGenericProcessor):

    def file_command_iter(self, paths, dir):

        # Revno 1: create multiple files
        # Revno 2: delete a directory holding those files
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(path, kind_to_mode('file', False),
                                                     None, b"aaa%d" % i)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(dir)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_dir(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d', b'a/b/e/f', b'a/g']
        dir = b'a/b'
        handler.process(self.file_command_iter(paths, dir))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[
                                                    (b'a',), (b'a/b',
                                                              ), (b'a/b/c',),
                                                    (b'a/b/d',),
                                                    (b'a/b/e',), (b'a/b/e/f',),
                                                    (b'a/g',),
                                                    ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a/b',), (b'a/b/c',),
                                                    (b'a/b/d',),
                                                    (b'a/b/e',), (b'a/b/e/f',),
                                                    ])


class TestImportToPackDeleteDirectoryThenAddFile(TestCaseForGenericProcessor):
    """Test deleting a directory then adding a file in the same commit."""

    def file_command_iter(self, paths, dir, new_path, kind='file'):

        # Revno 1: create files in a directory
        # Revno 2: delete the directory then add a file into it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(path, kind_to_mode(kind, False),
                                                     None, b"aaa%d" % i)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(dir)
                yield commands.FileModifyCommand(new_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_dir_then_add_file(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d']
        dir = b'a/b'
        new_path = b'a/b/z'
        handler.process(self.file_command_iter(paths, dir, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (b'a/b/c',), (b'a/b/d',), ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a/b',), (b'a/b/c',), (b'a/b/d',)],
                                                expected_added=[(b'a/b',), (b'a/b/z',)])
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_dir_then_add_symlink(self):
        handler, branch = self.get_handler()
        paths = [b'a/b/c', b'a/b/d']
        dir = b'a/b'
        new_path = b'a/b/z'
        handler.process(self.file_command_iter(
            paths, dir, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (b'a/b/c',), (b'a/b/d',), ])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'a/b',), (b'a/b/c',), (b'a/b/d',)],
                                                expected_added=[(b'a/b',), (b'a/b/z',)])
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRename(TestCaseForGenericProcessor):

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: rename it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileRenameCommand(old_path, new_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_rename_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'a/b'
        handler.process(self.get_command_iter(old_path, new_path))
        self.assertChanges(branch, 2, expected_renamed=[(old_path, new_path)])

    def test_rename_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'a/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        self.assertChanges(branch, 2, expected_renamed=[(old_path, new_path)])

    def test_rename_file_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'b/a'
        handler.process(self.get_command_iter(old_path, new_path))
        self.assertChanges(branch, 2,
                           expected_renamed=[(old_path, new_path)],
                           expected_added=[(b'b',)],
                           expected_removed=[(b'a',)])

    def test_rename_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'b/a'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        self.assertChanges(branch, 2,
                           expected_renamed=[(old_path, new_path)],
                           expected_added=[(b'b',)],
                           expected_removed=[(b'a',)])


class TestImportToPackRenameNew(TestCaseForGenericProcessor):
    """Test rename of a newly added file."""

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create a file and rename it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileRenameCommand(old_path, new_path)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)
        return command_list

    def test_rename_new_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(new_path,)])
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(new_path,)])
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'a/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (new_path,)])

    def test_rename_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'a/a'
        new_path = b'a/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (new_path,)])


class TestImportToPackRenameToDeleted(TestCaseForGenericProcessor):
    """Test rename to a destination path deleted in this commit."""

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create two files
        # Revno 2: delete one, rename the other one to that path
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileModifyCommand(new_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(new_path)
                yield commands.FileRenameCommand(old_path, new_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_rename_to_deleted_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_to_deleted_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_to_deleted_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")

    def test_rename_to_deleted_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")

    def test_rename_to_deleted_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,), (b'd2',), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'd1',), (new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")

    def test_rename_to_deleted_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,), (b'd2',), (new_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'd1',), (new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")


class TestImportToPackRenameModified(TestCaseForGenericProcessor):
    """Test rename of a path previously modified in this commit."""

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: modify then rename it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
                yield commands.FileRenameCommand(old_path, new_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_rename_of_modified_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_of_modified_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_of_modified_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_of_modified_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_rename_of_modified_file_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path, new_path)],
                                                expected_added=[(b'd2',)],
                                                expected_removed=[(b'd1',)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_of_modified_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path, new_path)],
                                                expected_added=[(b'd2',)],
                                                expected_removed=[(b'd1',)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRenameThenModify(TestCaseForGenericProcessor):
    """Test rename of a path then modfy the new-path in the same commit."""

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: rename it then modify the newly created path
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileRenameCommand(old_path, new_path)
                yield commands.FileModifyCommand(new_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_rename_then_modify_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_then_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_then_modify_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path, new_path)],
                                                expected_added=[(b'd2',)],
                                                expected_removed=[(b'd1',)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_then_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_then_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_rename_then_modify_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path, new_path)],
                                                expected_added=[(b'd2',)],
                                                expected_removed=[(b'd1',)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackDeleteRenameThenModify(TestCaseForGenericProcessor):
    """Test rename of to a deleted path then modfy the new-path in the same commit."""

    def get_command_iter(self, old_path, new_path, kind='file'):

        # Revno 1: create two files or symlinks
        # Revno 2: delete one, rename the other to it then modify the newly created path
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(old_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileModifyCommand(new_path, kind_to_mode(kind, False),
                                                 None, b"zzz")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(new_path)
                yield commands.FileRenameCommand(old_path, new_path)
                yield commands.FileModifyCommand(new_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_delete_rename_then_modify_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_delete_rename_then_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_rename_then_modify_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (b'd2',), (old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'd1',), (new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_rename_then_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b'a'
        new_path = b'b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_delete_rename_then_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b'd/a'
        new_path = b'd/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[(new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_delete_rename_then_modify_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b'd1/a'
        new_path = b'd2/b'
        handler.process(self.get_command_iter(old_path, new_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd1',), (b'd2',), (old_path,), (new_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (b'd1',), (new_path,)],
                                                expected_renamed=[(old_path, new_path)])
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRenameTricky(TestCaseForGenericProcessor):

    def file_command_iter(self, path1, old_path2, new_path2, kind='file'):

        # Revno 1: create two files or symlinks in a directory
        # Revno 2: rename the second file so that it implicitly deletes the
        # first one because either:
        # * the new file is a in directory with the old file name
        # * the new file has the same name as the directory of the first
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path1, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileModifyCommand(old_path2, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileRenameCommand(old_path2, new_path2)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_rename_file_becomes_directory(self):
        handler, branch = self.get_handler()
        old_path2 = b'foo'
        path1 = b'a/b'
        new_path2 = b'a/b/c'
        handler.process(self.file_command_iter(path1, old_path2, new_path2))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path1,), (old_path2,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path2, new_path2)],
                                                expected_kind_changed=[(path1, 'file', 'directory')])
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, new_path2, b"bbb")

    def test_rename_directory_becomes_file(self):
        handler, branch = self.get_handler()
        old_path2 = b'foo'
        path1 = b'a/b/c'
        new_path2 = b'a/b'
        handler.process(self.file_command_iter(path1, old_path2, new_path2))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (path1,), (old_path2,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path2, new_path2)],
                                                expected_removed=[(path1,), (new_path2,)])
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, new_path2, b"bbb")

    def test_rename_symlink_becomes_directory(self):
        handler, branch = self.get_handler()
        old_path2 = b'foo'
        path1 = b'a/b'
        new_path2 = b'a/b/c'
        handler.process(self.file_command_iter(path1, old_path2, new_path2,
                                               'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (path1,), (old_path2,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path2, new_path2)],
                                                expected_kind_changed=[(path1, 'symlink', 'directory')])
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path2, "bbb")

    def test_rename_directory_becomes_symlink(self):
        handler, branch = self.get_handler()
        old_path2 = b'foo'
        path1 = b'a/b/c'
        new_path2 = b'a/b'
        handler.process(self.file_command_iter(path1, old_path2, new_path2,
                                               'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (b'a/b',), (path1,), (old_path2,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_renamed=[
                                                    (old_path2, new_path2)],
                                                expected_removed=[(path1,), (new_path2,)])
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path2, "bbb")


class TestImportToPackCopy(TestCaseForGenericProcessor):

    def file_command_iter(self, src_path, dest_path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: copy it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(src_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileCopyCommand(src_path, dest_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_copy_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'a/b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'b/a'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(b'b',), (dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'a/b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")

    def test_copy_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'b/a'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_added=[(b'b',), (dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")


class TestImportToPackCopyNew(TestCaseForGenericProcessor):
    """Test copy of a newly added file."""

    def file_command_iter(self, src_path, dest_path, kind='file'):

        # Revno 1: create a file or symlink and copy it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(src_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileCopyCommand(src_path, dest_path)
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)
        return command_list

    def test_copy_new_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(src_path,), (dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'a/b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (src_path,), (dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")

    def test_copy_new_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'b/a'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (src_path,), (b'b',), (dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")

    def test_copy_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(src_path,), (dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'a/b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (src_path,), (dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")

    def test_copy_new_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'a/a'
        dest_path = b'b/a'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'a',), (src_path,), (b'b',), (dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")


class TestImportToPackCopyToDeleted(TestCaseForGenericProcessor):

    def file_command_iter(self, src_path, dest_path, kind='file'):

        # Revno 1: create two files or symlinks
        # Revno 2: delete one and copy the other one to its path
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(src_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
                yield commands.FileModifyCommand(dest_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileDeleteCommand(dest_path)
                yield commands.FileCopyCommand(src_path, dest_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_copy_to_deleted_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(src_path,), (dest_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (dest_path,)],
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"bbb")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_to_deleted_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(src_path,), (dest_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (dest_path,)],
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_to_deleted_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'd/a'
        dest_path = b'd/b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (src_path,), (dest_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (dest_path,)],
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"bbb")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_to_deleted_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'd/a'
        dest_path = b'd/b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree0, revtree1 = self.assertChanges(branch, 1,
                                                expected_added=[(b'd',), (src_path,), (dest_path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_removed=[
                                                    (dest_path,)],
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")


class TestImportToPackCopyModified(TestCaseForGenericProcessor):
    """Test copy of file/symlink already modified in this commit."""

    def file_command_iter(self, src_path, dest_path, kind='file'):

        # Revno 1: create a file or symlink
        # Revno 2: modify and copy it
        def command_list():
            author = [b'', b'bugs@a.com', time.time(), time.timezone]
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(src_path, kind_to_mode(kind, False),
                                                 None, b"aaa")
            yield commands.CommitCommand(b'head', b'1', author,
                                         committer, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(src_path, kind_to_mode(kind, False),
                                                 None, b"bbb")
                yield commands.FileCopyCommand(src_path, dest_path)
            yield commands.CommitCommand(b'head', b'2', author,
                                         committer, b"commit 2", b":1", [], files_two)
        return command_list

    def test_copy_of_modified_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_of_modified_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'd/a'
        dest_path = b'd/b'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")

    def test_copy_of_modified_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'd1/a'
        dest_path = b'd2/a'
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(b'd2',), (dest_path,)])
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")

    def test_copy_of_modified_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b'a'
        dest_path = b'b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_of_modified_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b'd/a'
        dest_path = b'd/b'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")

    def test_copy_of_modified_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b'd1/a'
        dest_path = b'd2/a'
        handler.process(self.file_command_iter(src_path, dest_path, 'symlink'))
        revtree1, revtree2 = self.assertChanges(branch, 2,
                                                expected_modified=[
                                                    (src_path,)],
                                                expected_added=[(b'd2',), (dest_path,)])
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")


class TestImportToPackFileKinds(TestCaseForGenericProcessor):

    def get_command_iter(self, path, kind, content):

        def command_list():
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, kind_to_mode(kind, False),
                                                 None, content)
            yield commands.CommitCommand(b'head', b'1', None,
                                         committer, b"commit 1", None, [], files_one)
        return command_list

    def test_import_plainfile(self):
        handler, branch = self.get_handler()
        handler.process(self.get_command_iter(b'foo', 'file', b'aaa'))

    def test_import_symlink(self):
        handler, branch = self.get_handler()
        handler.process(self.get_command_iter(b'foo', 'symlink', b'bar'))


class TestModifyRevertInBranch(TestCaseForGenericProcessor):

    def file_command_iter(self):
        # A     add 'foo'
        # |\
        # | B   modify 'foo'
        # | |
        # | C   revert 'foo' back to A
        # |/
        # D     merge 'foo'
        def command_list():
            committer_a = [b'', b'a@elmer.com', time.time(), time.timezone]
            committer_b = [b'', b'b@elmer.com', time.time(), time.timezone]
            committer_c = [b'', b'c@elmer.com', time.time(), time.timezone]
            committer_d = [b'', b'd@elmer.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(b'foo', kind_to_mode('file', False),
                                                 None, b"content A\n")
            yield commands.CommitCommand(b'head', b'1', None,
                                         committer_a, b"commit 1", None, [], files_one)

            def files_two():
                yield commands.FileModifyCommand(b'foo', kind_to_mode('file', False),
                                                 None, b"content B\n")
            yield commands.CommitCommand(b'head', b'2', None,
                                         committer_b, b"commit 2", b":1", [], files_two)

            def files_three():
                yield commands.FileModifyCommand(b'foo', kind_to_mode('file', False),
                                                 None, b"content A\n")
            yield commands.CommitCommand(b'head', b'3', None,
                                         committer_c, b"commit 3", b":2", [], files_three)
            yield commands.CommitCommand(b'head', b'4', None,
                                         committer_d, b"commit 4", b":1", [b':3'], lambda: [])
        return command_list

    def test_modify_revert(self):
        handler, branch = self.get_handler()
        handler.process(self.file_command_iter())
        branch.lock_read()
        self.addCleanup(branch.unlock)
        rev_d = branch.last_revision()
        rev_a, rev_c = branch.repository.get_parent_map([rev_d])[rev_d]
        rev_b = branch.repository.get_parent_map([rev_c])[rev_c][0]
        rtree_a, rtree_b, rtree_c, rtree_d = branch.repository.revision_trees([
            rev_a, rev_b, rev_c, rev_d])
        self.assertEqual(rev_a, rtree_a.get_file_revision('foo'))
        self.assertEqual(rev_b, rtree_b.get_file_revision('foo'))
        self.assertEqual(rev_c, rtree_c.get_file_revision('foo'))
        self.assertEqual(rev_c, rtree_d.get_file_revision('foo'))


class TestCommitCommands(TestCaseForGenericProcessor):

    def test_non_utf8_commit_message(self):
        handler, branch = self.get_handler()

        def files_one():
            yield commands.FileModifyCommand(b'a',
                                             kind_to_mode('file', False), None, b"data")

        def command_list():
            committer = [b'', b'elmer@a.com', time.time(), time.timezone]
            yield commands.CommitCommand(b'head', b'1', None,
                                         committer, b'This is a funky character: \x83', None, [],
                                         files_one)
        handler.process(command_list)
        rev = branch.repository.get_revision(branch.last_revision())
        self.assertEqual("This is a funky character: \ufffd", rev.message)


class TestAddNonUtf8InBranch(TestCaseForGenericProcessor):

    def file_command_iter(self):
        # A     add 'foo\x83'
        def command_list():
            committer_a = [b'', b'a@elmer.com', time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    b'foo\x83', kind_to_mode('file', False), None, b"content A\n")
            yield commands.CommitCommand(b'head', b'1', None,
                                         committer_a, b"commit 1", None, [], files_one)
        return command_list

    def test_add(self):
        handler, branch = self.get_handler()
        handler.process(self.file_command_iter())
        branch.lock_read()
        self.addCleanup(branch.unlock)
        rev_a = branch.last_revision()
        rtree_a = branch.repository.revision_tree(rev_a)
        self.assertEqual(rev_a, rtree_a.get_file_revision('foo\ufffd'))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests that use BrokenRepoScenario objects.

That is, tests for reconcile and check.
"""

from breezy import osutils
from breezy.bzr.inventory import Inventory, InventoryFile
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository, all_repository_vf_format_scenarios)
from breezy.repository import WriteGroup
from breezy.revision import NULL_REVISION, Revision
from breezy.tests import TestNotApplicable, multiply_scenarios
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class BrokenRepoScenario:
    """Base class for defining scenarios for testing check and reconcile.

    A subclass needs to define the following methods:
        :populate_repository: a method to use to populate a repository with
            sample revisions, inventories and file versions.
        :all_versions_after_reconcile: all the versions in repository after
            reconcile.  run_test verifies that the text of each of these
            versions of the file is unchanged by the reconcile.
        :populated_parents: a list of (parents list, revision).  Each version
            of the file is verified to have the given parents before running
            the reconcile.  i.e. this is used to assert that the repo from the
            factory is what we expect.
        :corrected_parents: a list of (parents list, revision).  Each version
            of the file is verified to have the given parents after the
            reconcile.  i.e. this is used to assert that reconcile made the
            changes we expect it to make.

    A subclass may define the following optional method as well:
        :corrected_fulltexts: a list of file versions that should be stored as
            fulltexts (not deltas) after reconcile.  run_test will verify that
            this occurs.
    """

    def __init__(self, test_case):
        self.test_case = test_case

    def make_one_file_inventory(self, repo, revision, parents,
                                inv_revision=None, root_revision=None,
                                file_contents=None, make_file_version=True):
        return self.test_case.make_one_file_inventory(
            repo, revision, parents, inv_revision=inv_revision,
            root_revision=root_revision, file_contents=file_contents,
            make_file_version=make_file_version)

    def add_revision(self, repo, revision_id, inv, parent_ids):
        return self.test_case.add_revision(repo, revision_id, inv, parent_ids)

    def corrected_fulltexts(self):
        return []

    def repository_text_key_index(self):
        result = {}
        if self.versioned_root:
            result.update(self.versioned_repository_text_keys())
        result.update(self.repository_text_keys())
        return result


class UndamagedRepositoryScenario(BrokenRepoScenario):
    """A scenario where the repository has no damage.

    It has a single revision, 'rev1a', with a single file.
    """

    def all_versions_after_reconcile(self):
        return (b'rev1a', )

    def populated_parents(self):
        return (((), b'rev1a'), )

    def corrected_parents(self):
        # Same as the populated parents, because there was nothing wrong.
        return self.populated_parents()

    def check_regexes(self, repo):
        return ["0 unreferenced text versions"]

    def populate_repository(self, repo):
        # make rev1a: A well-formed revision, containing 'a-file'
        inv = self.make_one_file_inventory(
            repo, b'rev1a', [], root_revision=b'rev1a')
        self.add_revision(repo, b'rev1a', inv, [])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'rev1a'): True})
        result.update({(b'a-file-id', b'rev1a'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'rev1a'): [NULL_REVISION]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'rev1a'): [NULL_REVISION]}


class FileParentIsNotInRevisionAncestryScenario(BrokenRepoScenario):
    """A scenario where a revision 'rev2' has 'a-file' with a
    parent 'rev1b' that is not in the revision ancestry.

    Reconcile should remove 'rev1b' from the parents list of 'a-file' in
    'rev2', preserving 'rev1a' as a parent.
    """

    def all_versions_after_reconcile(self):
        return (b'rev1a', b'rev2')

    def populated_parents(self):
        return (
            ((), b'rev1a'),
            ((), b'rev1b'),  # Will be gc'd
            ((b'rev1a', b'rev1b'), b'rev2'))  # Will have parents trimmed

    def corrected_parents(self):
        return (
            ((), b'rev1a'),
            (None, b'rev1b'),
            ((b'rev1a',), b'rev2'))

    def check_regexes(self, repo):
        return [r"\* a-file-id version rev2 has parents \(rev1a, rev1b\) "
                r"but should have \(rev1a\)",
                "1 unreferenced text versions",
                ]

    def populate_repository(self, repo):
        # make rev1a: A well-formed revision, containing 'a-file'
        inv = self.make_one_file_inventory(
            repo, b'rev1a', [], root_revision=b'rev1a')
        self.add_revision(repo, b'rev1a', inv, [])

        # make rev1b, which has no Revision, but has an Inventory, and
        # a-file
        inv = self.make_one_file_inventory(
            repo, b'rev1b', [], root_revision=b'rev1b')
        repo.add_inventory(b'rev1b', inv, [])

        # make rev2, with a-file.
        # a-file has 'rev1b' as an ancestor, even though this is not
        # mentioned by 'rev1a', making it an unreferenced ancestor
        inv = self.make_one_file_inventory(
            repo, b'rev2', [b'rev1a', b'rev1b'])
        self.add_revision(repo, b'rev2', inv, [b'rev1a'])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'rev1a'): True,
                           (b'TREE_ROOT', b'rev2'): True})
        result.update({(b'a-file-id', b'rev1a'): True,
                       (b'a-file-id', b'rev2'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'rev1a'): [NULL_REVISION],
                (b'a-file-id', b'rev2'): [(b'a-file-id', b'rev1a')]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'rev1a'): [NULL_REVISION],
                (b'TREE_ROOT', b'rev2'): [(b'TREE_ROOT', b'rev1a')]}


class FileParentHasInaccessibleInventoryScenario(BrokenRepoScenario):
    """A scenario where a revision 'rev3' containing 'a-file' modified in
    'rev3', and with a parent which is in the revision ancestory, but whose
    inventory cannot be accessed at all.

    Reconcile should remove the file version parent whose inventory is
    inaccessbile (i.e. remove 'rev1c' from the parents of a-file's rev3).
    """

    def all_versions_after_reconcile(self):
        return (b'rev2', b'rev3')

    def populated_parents(self):
        return (
            ((), b'rev2'),
            ((b'rev1c',), b'rev3'))

    def corrected_parents(self):
        return (
            ((), b'rev2'),
            ((), b'rev3'))

    def check_regexes(self, repo):
        return [r"\* a-file-id version rev3 has parents "
                r"\(rev1c\) but should have \(\)",
                ]

    def populate_repository(self, repo):
        # make rev2, with a-file
        # a-file is sane
        inv = self.make_one_file_inventory(repo, b'rev2', [])
        self.add_revision(repo, b'rev2', inv, [])

        # make ghost revision rev1c, with a version of a-file present so
        # that we generate a knit delta against this version.  In real life
        # the ghost might never have been present or rev3 might have been
        # generated against a revision that was present at the time.  So
        # currently we have the full history of a-file present even though
        # the inventory and revision objects are not.
        self.make_one_file_inventory(repo, b'rev1c', [])

        # make rev3 with a-file
        # a-file refers to 'rev1c', which is a ghost in this repository, so
        # a-file cannot have rev1c as its ancestor.
        inv = self.make_one_file_inventory(repo, b'rev3', [b'rev1c'])
        self.add_revision(repo, b'rev3', inv, [b'rev1c', b'rev1a'])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'rev2'): True,
                           (b'TREE_ROOT', b'rev3'): True})
        result.update({(b'a-file-id', b'rev2'): True,
                       (b'a-file-id', b'rev3'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'rev2'): [NULL_REVISION],
                (b'a-file-id', b'rev3'): [NULL_REVISION]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'rev2'): [NULL_REVISION],
                (b'TREE_ROOT', b'rev3'): [NULL_REVISION]}


class FileParentsNotReferencedByAnyInventoryScenario(BrokenRepoScenario):
    """A scenario where a repository with file 'a-file' which has extra
    per-file versions that are not referenced by any inventory (even though
    they have the same ID as actual revisions).  The inventory of 'rev2'
    references 'rev1a' of 'a-file', but there is a 'rev2' of 'some-file' stored
    and erroneously referenced by later per-file versions (revisions 'rev4' and
    'rev5').

    Reconcile should remove the file parents that are not referenced by any
    inventory.
    """

    def all_versions_after_reconcile(self):
        return (b'rev1a', b'rev2c', b'rev4', b'rev5')

    def populated_parents(self):
        return [
            ((b'rev1a',), b'rev2'),
            ((b'rev1a',), b'rev2b'),
            ((b'rev2',), b'rev3'),
            ((b'rev2',), b'rev4'),
            ((b'rev2', b'rev2c'), b'rev5')]

    def corrected_parents(self):
        return (
            # rev2 and rev2b have been removed.
            (None, b'rev2'),
            (None, b'rev2b'),
            # rev3's accessible parent inventories all have rev1a as the last
            # modifier.
            ((b'rev1a',), b'rev3'),
            # rev1a features in both rev4's parents but should only appear once
            # in the result
            ((b'rev1a',), b'rev4'),
            # rev2c is the head of rev1a and rev2c, the inventory provided
            # per-file last-modified revisions.
            ((b'rev2c',), b'rev5'))

    def check_regexes(self, repo):
        if repo.supports_rich_root():
            # TREE_ROOT will be wrong; but we're not testing it. so just adjust
            # the expected count of errors.
            count = 9
        else:
            count = 3
        return [
            # will be gc'd
            r"unreferenced version: {rev2} in a-file-id",
            r"unreferenced version: {rev2b} in a-file-id",
            # will be corrected
            r"a-file-id version rev3 has parents \(rev2\) "
            r"but should have \(rev1a\)",
            r"a-file-id version rev5 has parents \(rev2, rev2c\) "
            r"but should have \(rev2c\)",
            r"a-file-id version rev4 has parents \(rev2\) "
            r"but should have \(rev1a\)",
            "%d inconsistent parents" % count,
            ]

    def populate_repository(self, repo):
        # make rev1a: A well-formed revision, containing 'a-file'
        inv = self.make_one_file_inventory(
            repo, b'rev1a', [], root_revision=b'rev1a')
        self.add_revision(repo, b'rev1a', inv, [])

        # make rev2, with a-file.
        # a-file is unmodified from rev1a, and an unreferenced rev2 file
        # version is present in the repository.
        self.make_one_file_inventory(
            repo, b'rev2', [b'rev1a'], inv_revision=b'rev1a')
        self.add_revision(repo, b'rev2', inv, [b'rev1a'])

        # make rev3 with a-file
        # a-file has 'rev2' as its ancestor, but the revision in 'rev2' was
        # rev1a so this is inconsistent with rev2's inventory - it should
        # be rev1a, and at the revision level 1c is not present - it is a
        # ghost, so only the details from rev1a are available for
        # determining whether a delta is acceptable, or a full is needed,
        # and what the correct parents are.
        inv = self.make_one_file_inventory(repo, b'rev3', [b'rev2'])
        self.add_revision(repo, b'rev3', inv, [b'rev1c', b'rev1a'])

        # In rev2b, the true last-modifying-revision of a-file is rev1a,
        # inherited from rev2, but there is a version rev2b of the file, which
        # reconcile could remove, leaving no rev2b.  Most importantly,
        # revisions descending from rev2b should not have per-file parents of
        # a-file-rev2b.
        # ??? This is to test deduplication in fixing rev4
        inv = self.make_one_file_inventory(
            repo, b'rev2b', [b'rev1a'], inv_revision=b'rev1a')
        self.add_revision(repo, b'rev2b', inv, [b'rev1a'])

        # rev4 is for testing that when the last modified of a file in
        # multiple parent revisions is the same, that it only appears once
        # in the generated per file parents list: rev2 and rev2b both
        # descend from 1a and do not change the file a-file, so there should
        # be no version of a-file 'rev2' or 'rev2b', but rev4 does change
        # a-file, and is a merge of rev2 and rev2b, so it should end up with
        # a parent of just rev1a - the starting file parents list is simply
        # completely wrong.
        inv = self.make_one_file_inventory(repo, b'rev4', [b'rev2'])
        self.add_revision(repo, b'rev4', inv, [b'rev2', b'rev2b'])

        # rev2c changes a-file from rev1a, so the version it of a-file it
        # introduces is a head revision when rev5 is checked.
        inv = self.make_one_file_inventory(repo, b'rev2c', [b'rev1a'])
        self.add_revision(repo, b'rev2c', inv, [b'rev1a'])

        # rev5 descends from rev2 and rev2c; as rev2 does not alter a-file,
        # but rev2c does, this should use rev2c as the parent for the per
        # file history, even though more than one per-file parent is
        # available, because we use the heads of the revision parents for
        # the inventory modification revisions of the file to determine the
        # parents for the per file graph.
        inv = self.make_one_file_inventory(repo, b'rev5', [b'rev2', b'rev2c'])
        self.add_revision(repo, b'rev5', inv, [b'rev2', b'rev2c'])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'rev1a'): True,
                           (b'TREE_ROOT', b'rev2'): True,
                           (b'TREE_ROOT', b'rev2b'): True,
                           (b'TREE_ROOT', b'rev2c'): True,
                           (b'TREE_ROOT', b'rev3'): True,
                           (b'TREE_ROOT', b'rev4'): True,
                           (b'TREE_ROOT', b'rev5'): True})
        result.update({(b'a-file-id', b'rev1a'): True,
                       (b'a-file-id', b'rev2c'): True,
                       (b'a-file-id', b'rev3'): True,
                       (b'a-file-id', b'rev4'): True,
                       (b'a-file-id', b'rev5'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'rev1a'): [NULL_REVISION],
                (b'a-file-id', b'rev2c'): [(b'a-file-id', b'rev1a')],
                (b'a-file-id', b'rev3'): [(b'a-file-id', b'rev1a')],
                (b'a-file-id', b'rev4'): [(b'a-file-id', b'rev1a')],
                (b'a-file-id', b'rev5'): [(b'a-file-id', b'rev2c')]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'rev1a'): [NULL_REVISION],
                (b'TREE_ROOT', b'rev2'): [(b'TREE_ROOT', b'rev1a')],
                (b'TREE_ROOT', b'rev2b'): [(b'TREE_ROOT', b'rev1a')],
                (b'TREE_ROOT', b'rev2c'): [(b'TREE_ROOT', b'rev1a')],
                (b'TREE_ROOT', b'rev3'): [(b'TREE_ROOT', b'rev1a')],
                (b'TREE_ROOT', b'rev4'):
                    [(b'TREE_ROOT', b'rev2'), (b'TREE_ROOT', b'rev2b')],
                (b'TREE_ROOT', b'rev5'):
                    [(b'TREE_ROOT', b'rev2'), (b'TREE_ROOT', b'rev2c')]}


class UnreferencedFileParentsFromNoOpMergeScenario(BrokenRepoScenario):
    """
    rev1a and rev1b with identical contents
    rev2 revision has parents of [rev1a, rev1b]
    There is a a-file:rev2 file version, not referenced by the inventory.
    """

    def all_versions_after_reconcile(self):
        return (b'rev1a', b'rev1b', b'rev2', b'rev4')

    def populated_parents(self):
        return (
            ((), b'rev1a'),
            ((), b'rev1b'),
            ((b'rev1a', b'rev1b'), b'rev2'),
            (None, b'rev3'),
            ((b'rev2',), b'rev4'),
            )

    def corrected_parents(self):
        return (
            ((), b'rev1a'),
            ((), b'rev1b'),
            ((), b'rev2'),
            (None, b'rev3'),
            ((b'rev2',), b'rev4'),
            )

    def corrected_fulltexts(self):
        return [b'rev2']

    def check_regexes(self, repo):
        return []

    def populate_repository(self, repo):
        # make rev1a: A well-formed revision, containing 'a-file'
        inv1a = self.make_one_file_inventory(
            repo, b'rev1a', [], root_revision=b'rev1a')
        self.add_revision(repo, b'rev1a', inv1a, [])

        # make rev1b: A well-formed revision, containing 'a-file'
        # rev1b of a-file has the exact same contents as rev1a.
        file_contents = next(
            repo.texts.get_record_stream([(b'a-file-id', b'rev1a')],
                                         "unordered", False)).get_bytes_as('fulltext')
        inv = self.make_one_file_inventory(
            repo, b'rev1b', [], root_revision=b'rev1b',
            file_contents=file_contents)
        self.add_revision(repo, b'rev1b', inv, [])

        # make rev2, a merge of rev1a and rev1b, with a-file.
        # a-file is unmodified from rev1a and rev1b, but a new version is
        # wrongly present anyway.
        inv = self.make_one_file_inventory(
            repo, b'rev2', [b'rev1a', b'rev1b'], inv_revision=b'rev1a',
            file_contents=file_contents)
        self.add_revision(repo, b'rev2', inv, [b'rev1a', b'rev1b'])

        # rev3: a-file unchanged from rev2, but wrongly referencing rev2 of the
        # file in its inventory.
        inv = self.make_one_file_inventory(
            repo, b'rev3', [b'rev2'], inv_revision=b'rev2',
            file_contents=file_contents, make_file_version=False)
        self.add_revision(repo, b'rev3', inv, [b'rev2'])

        # rev4: a modification of a-file on top of rev3.
        inv = self.make_one_file_inventory(repo, b'rev4', [b'rev2'])
        self.add_revision(repo, b'rev4', inv, [b'rev3'])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'rev1a'): True,
                           (b'TREE_ROOT', b'rev1b'): True,
                           (b'TREE_ROOT', b'rev2'): True,
                           (b'TREE_ROOT', b'rev3'): True,
                           (b'TREE_ROOT', b'rev4'): True})
        result.update({(b'a-file-id', b'rev1a'): True,
                       (b'a-file-id', b'rev1b'): True,
                       (b'a-file-id', b'rev2'): False,
                       (b'a-file-id', b'rev4'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'rev1a'): [NULL_REVISION],
                (b'a-file-id', b'rev1b'): [NULL_REVISION],
                (b'a-file-id', b'rev2'): [NULL_REVISION],
                (b'a-file-id', b'rev4'): [(b'a-file-id', b'rev2')]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'rev1a'): [NULL_REVISION],
                (b'TREE_ROOT', b'rev1b'): [NULL_REVISION],
                (b'TREE_ROOT', b'rev2'):
                    [(b'TREE_ROOT', b'rev1a'), (b'TREE_ROOT', b'rev1b')],
                (b'TREE_ROOT', b'rev3'): [(b'TREE_ROOT', b'rev2')],
                (b'TREE_ROOT', b'rev4'): [(b'TREE_ROOT', b'rev3')]}


class TooManyParentsScenario(BrokenRepoScenario):
    """A scenario where 'broken-revision' of 'a-file' claims to have parents
    ['good-parent', 'bad-parent'].  However 'bad-parent' is in the ancestry of
    'good-parent', so the correct parent list for that file version are is just
    ['good-parent'].
    """

    def all_versions_after_reconcile(self):
        return (b'bad-parent', b'good-parent', b'broken-revision')

    def populated_parents(self):
        return (
            ((), b'bad-parent'),
            ((b'bad-parent',), b'good-parent'),
            ((b'good-parent', b'bad-parent'), b'broken-revision'))

    def corrected_parents(self):
        return (
            ((), b'bad-parent'),
            ((b'bad-parent',), b'good-parent'),
            ((b'good-parent',), b'broken-revision'))

    def check_regexes(self, repo):
        if repo.supports_rich_root():
            # TREE_ROOT will be wrong; but we're not testing it. so just adjust
            # the expected count of errors.
            count = 3
        else:
            count = 1
        return (
            '     %d inconsistent parents' % count,
            (r"      \* a-file-id version broken-revision has parents "
             r"\(good-parent, bad-parent\) but "
             r"should have \(good-parent\)"))

    def populate_repository(self, repo):
        inv = self.make_one_file_inventory(
            repo, b'bad-parent', (), root_revision=b'bad-parent')
        self.add_revision(repo, b'bad-parent', inv, ())

        inv = self.make_one_file_inventory(
            repo, b'good-parent', (b'bad-parent',))
        self.add_revision(repo, b'good-parent', inv, (b'bad-parent',))

        inv = self.make_one_file_inventory(
            repo, b'broken-revision', (b'good-parent', b'bad-parent'))
        self.add_revision(repo, b'broken-revision', inv, (b'good-parent',))
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'bad-parent'): True,
                           (b'TREE_ROOT', b'broken-revision'): True,
                           (b'TREE_ROOT', b'good-parent'): True})
        result.update({(b'a-file-id', b'bad-parent'): True,
                       (b'a-file-id', b'broken-revision'): True,
                       (b'a-file-id', b'good-parent'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'bad-parent'): [NULL_REVISION],
                (b'a-file-id', b'broken-revision'):
                    [(b'a-file-id', b'good-parent')],
                (b'a-file-id', b'good-parent'): [(b'a-file-id', b'bad-parent')]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'bad-parent'): [NULL_REVISION],
                (b'TREE_ROOT', b'broken-revision'):
                    [(b'TREE_ROOT', b'good-parent')],
                (b'TREE_ROOT', b'good-parent'): [(b'TREE_ROOT', b'bad-parent')]}


class ClaimedFileParentDidNotModifyFileScenario(BrokenRepoScenario):
    """A scenario where the file parent is the same as the revision parent, but
    should not be because that revision did not modify the file.

    Specifically, the parent revision of 'current' is
    'modified-something-else', which does not modify 'a-file', but the
    'current' version of 'a-file' erroneously claims that
    'modified-something-else' is the parent file version.
    """

    def all_versions_after_reconcile(self):
        return (b'basis', b'current')

    def populated_parents(self):
        return (
            ((), b'basis'),
            ((b'basis',), b'modified-something-else'),
            ((b'modified-something-else',), b'current'))

    def corrected_parents(self):
        return (
            ((), b'basis'),
            (None, b'modified-something-else'),
            ((b'basis',), b'current'))

    def check_regexes(self, repo):
        if repo.supports_rich_root():
            # TREE_ROOT will be wrong; but we're not testing it. so just adjust
            # the expected count of errors.
            count = 3
        else:
            count = 1
        return (
            "%d inconsistent parents" % count,
            r"\* a-file-id version current has parents "
            r"\(modified-something-else\) but should have \(basis\)",
            )

    def populate_repository(self, repo):
        inv = self.make_one_file_inventory(repo, b'basis', ())
        self.add_revision(repo, b'basis', inv, ())

        # 'modified-something-else' is a correctly recorded revision, but it
        # does not modify the file we are looking at, so the inventory for that
        # file in this revision points to 'basis'.
        inv = self.make_one_file_inventory(
            repo, b'modified-something-else', (b'basis',), inv_revision=b'basis')
        self.add_revision(repo, b'modified-something-else', inv, (b'basis',))

        # The 'current' revision has 'modified-something-else' as its parent,
        # but the 'current' version of 'a-file' should have 'basis' as its
        # parent.
        inv = self.make_one_file_inventory(
            repo, b'current', (b'modified-something-else',))
        self.add_revision(repo, b'current', inv, (b'modified-something-else',))
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'basis'): True,
                           (b'TREE_ROOT', b'current'): True,
                           (b'TREE_ROOT', b'modified-something-else'): True})
        result.update({(b'a-file-id', b'basis'): True,
                       (b'a-file-id', b'current'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'basis'): [NULL_REVISION],
                (b'a-file-id', b'current'): [(b'a-file-id', b'basis')]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'basis'): [b'null:'],
                (b'TREE_ROOT', b'current'):
                    [(b'TREE_ROOT', b'modified-something-else')],
                (b'TREE_ROOT', b'modified-something-else'):
                    [(b'TREE_ROOT', b'basis')]}


class IncorrectlyOrderedParentsScenario(BrokenRepoScenario):
    """A scenario where the set parents of a version of a file are correct, but
    the order of those parents is incorrect.

    This defines a 'broken-revision-1-2' and a 'broken-revision-2-1' which both
    have their file version parents reversed compared to the revision parents,
    which is invalid.  (We use two revisions with opposite orderings of the
    same parents to make sure that accidentally relying on dictionary/set
    ordering cannot make the test pass; the assumption is that while dict/set
    iteration order is arbitrary, it is also consistent within a single test).
    """

    def all_versions_after_reconcile(self):
        return [b'parent-1', b'parent-2', b'broken-revision-1-2',
                b'broken-revision-2-1']

    def populated_parents(self):
        return (
            ((), b'parent-1'),
            ((), b'parent-2'),
            ((b'parent-2', b'parent-1'), b'broken-revision-1-2'),
            ((b'parent-1', b'parent-2'), b'broken-revision-2-1'))

    def corrected_parents(self):
        return (
            ((), b'parent-1'),
            ((), b'parent-2'),
            ((b'parent-1', b'parent-2'), b'broken-revision-1-2'),
            ((b'parent-2', b'parent-1'), b'broken-revision-2-1'))

    def check_regexes(self, repo):
        if repo.supports_rich_root():
            # TREE_ROOT will be wrong; but we're not testing it. so just adjust
            # the expected count of errors.
            count = 4
        else:
            count = 2
        return (
            "%d inconsistent parents" % count,
            r"\* a-file-id version broken-revision-1-2 has parents "
            r"\(parent-2, parent-1\) but should have "
            r"\(parent-1, parent-2\)",
            r"\* a-file-id version broken-revision-2-1 has parents "
            r"\(parent-1, parent-2\) but should have "
            r"\(parent-2, parent-1\)")

    def populate_repository(self, repo):
        inv = self.make_one_file_inventory(repo, b'parent-1', [])
        self.add_revision(repo, b'parent-1', inv, [])

        inv = self.make_one_file_inventory(repo, b'parent-2', [])
        self.add_revision(repo, b'parent-2', inv, [])

        inv = self.make_one_file_inventory(
            repo, b'broken-revision-1-2', [b'parent-2', b'parent-1'])
        self.add_revision(
            repo, b'broken-revision-1-2', inv, [b'parent-1', b'parent-2'])

        inv = self.make_one_file_inventory(
            repo, b'broken-revision-2-1', [b'parent-1', b'parent-2'])
        self.add_revision(
            repo, b'broken-revision-2-1', inv, [b'parent-2', b'parent-1'])
        self.versioned_root = repo.supports_rich_root()

    def repository_text_key_references(self):
        result = {}
        if self.versioned_root:
            result.update({(b'TREE_ROOT', b'broken-revision-1-2'): True,
                           (b'TREE_ROOT', b'broken-revision-2-1'): True,
                           (b'TREE_ROOT', b'parent-1'): True,
                           (b'TREE_ROOT', b'parent-2'): True})
        result.update({(b'a-file-id', b'broken-revision-1-2'): True,
                       (b'a-file-id', b'broken-revision-2-1'): True,
                       (b'a-file-id', b'parent-1'): True,
                       (b'a-file-id', b'parent-2'): True})
        return result

    def repository_text_keys(self):
        return {(b'a-file-id', b'broken-revision-1-2'):
                [(b'a-file-id', b'parent-1'), (b'a-file-id', b'parent-2')],
                (b'a-file-id', b'broken-revision-2-1'):
                    [(b'a-file-id', b'parent-2'), (b'a-file-id', b'parent-1')],
                (b'a-file-id', b'parent-1'): [NULL_REVISION],
                (b'a-file-id', b'parent-2'): [NULL_REVISION]}

    def versioned_repository_text_keys(self):
        return {(b'TREE_ROOT', b'broken-revision-1-2'):
                [(b'TREE_ROOT', b'parent-1'), (b'TREE_ROOT', b'parent-2')],
                (b'TREE_ROOT', b'broken-revision-2-1'):
                    [(b'TREE_ROOT', b'parent-2'), (b'TREE_ROOT', b'parent-1')],
                (b'TREE_ROOT', b'parent-1'): [NULL_REVISION],
                (b'TREE_ROOT', b'parent-2'): [NULL_REVISION]}


all_broken_scenario_classes = [
    UndamagedRepositoryScenario,
    FileParentIsNotInRevisionAncestryScenario,
    FileParentHasInaccessibleInventoryScenario,
    FileParentsNotReferencedByAnyInventoryScenario,
    TooManyParentsScenario,
    ClaimedFileParentDidNotModifyFileScenario,
    IncorrectlyOrderedParentsScenario,
    UnreferencedFileParentsFromNoOpMergeScenario,
    ]


def broken_scenarios_for_all_formats():
    format_scenarios = all_repository_vf_format_scenarios()
    # test_check_reconcile needs to be parameterized by format *and* by broken
    # repository scenario.
    broken_scenarios = [(s.__name__, {'scenario_class': s})
                        for s in all_broken_scenario_classes]
    return multiply_scenarios(format_scenarios, broken_scenarios)


class TestFileParentReconciliation(TestCaseWithRepository):
    """Tests for how reconcile corrects errors in parents of file versions."""

    scenarios = broken_scenarios_for_all_formats()

    def make_populated_repository(self, factory):
        """Create a new repository populated by the given factory."""
        repo = self.make_repository('broken-repo')
        with repo.lock_write(), WriteGroup(repo):
            factory(repo)
            return repo

    def add_revision(self, repo, revision_id, inv, parent_ids):
        """Add a revision with a given inventory and parents to a repository.

        :param repo: a repository.
        :param revision_id: the revision ID for the new revision.
        :param inv: an inventory (such as created by
            `make_one_file_inventory`).
        :param parent_ids: the parents for the new revision.
        """
        inv.revision_id = revision_id
        inv.root.revision = revision_id
        if repo.supports_rich_root():
            root_id = inv.root.file_id
            repo.texts.add_lines((root_id, revision_id), [], [])
        repo.add_inventory(revision_id, inv, parent_ids)
        revision = Revision(revision_id, committer='jrandom@example.com',
                            timestamp=0, inventory_sha1='', timezone=0, message='foo',
                            parent_ids=parent_ids)
        repo.add_revision(revision_id, revision, inv)

    def make_one_file_inventory(self, repo, revision, parents,
                                inv_revision=None, root_revision=None,
                                file_contents=None, make_file_version=True):
        """Make an inventory containing a version of a file with ID 'a-file'.

        The file's ID will be 'a-file', and its filename will be 'a file name',
        stored at the tree root.

        :param repo: a repository to add the new file version to.
        :param revision: the revision ID of the new inventory.
        :param parents: the parents for this revision of 'a-file'.
        :param inv_revision: if not None, the revision ID to store in the
            inventory entry.  Otherwise, this defaults to revision.
        :param root_revision: if not None, the inventory's root.revision will
            be set to this.
        :param file_contents: if not None, the contents of this file version.
            Otherwise a unique default (based on revision ID) will be
            generated.
        """
        inv = Inventory(revision_id=revision)
        if root_revision is not None:
            inv.root.revision = root_revision
        file_id = b'a-file-id'
        entry = InventoryFile(file_id, 'a file name', b'TREE_ROOT')
        if inv_revision is not None:
            entry.revision = inv_revision
        else:
            entry.revision = revision
        entry.text_size = 0
        if file_contents is None:
            file_contents = b'%sline\n' % entry.revision
        entry.text_sha1 = osutils.sha_string(file_contents)
        inv.add(entry)
        if make_file_version:
            repo.texts.add_lines((file_id, revision),
                                 [(file_id, parent) for parent in parents], [file_contents])
        return inv

    def require_repo_suffers_text_parent_corruption(self, repo):
        if not repo._reconcile_fixes_text_parents:
            raise TestNotApplicable(
                "Format does not support text parent reconciliation")

    def file_parents(self, repo, revision_id):
        key = (b'a-file-id', revision_id)
        parent_map = repo.texts.get_parent_map([key])
        return tuple(parent[-1] for parent in parent_map[key])

    def assertFileVersionAbsent(self, repo, revision_id):
        self.assertEqual({},
                         repo.texts.get_parent_map([(b'a-file-id', revision_id)]))

    def assertParentsMatch(self, expected_parents_for_versions, repo,
                           when_description):
        for expected_parents, version in expected_parents_for_versions:
            if expected_parents is None:
                self.assertFileVersionAbsent(repo, version)
            else:
                found_parents = self.file_parents(repo, version)
                self.assertEqual(expected_parents, found_parents,
                                 "%s reconcile %s has parents %s, should have %s."
                                 % (when_description, version, found_parents,
                                    expected_parents))

    def prepare_test_repository(self):
        """Prepare a repository to test with from the test scenario.

        :return: A repository, and the scenario instance.
        """
        scenario = self.scenario_class(self)
        repo = self.make_populated_repository(scenario.populate_repository)
        self.require_repo_suffers_text_parent_corruption(repo)
        return repo, scenario

    def shas_for_versions_of_file(self, repo, versions):
        """Get the SHA-1 hashes of the versions of 'a-file' in the repository.

        :param repo: the repository to get the hashes from.
        :param versions: a list of versions to get hashes for.

        :returns: A dict of `{version: hash}`.
        """
        keys = [(b'a-file-id', version) for version in versions]
        return repo.texts.get_sha1s(keys)

    def test_reconcile_behaviour(self):
        """Populate a repository and reconcile it, verifying the state before
        and after.
        """
        repo, scenario = self.prepare_test_repository()
        with repo.lock_read():
            self.assertParentsMatch(scenario.populated_parents(), repo,
                                    b'before')
            vf_shas = self.shas_for_versions_of_file(
                repo, scenario.all_versions_after_reconcile())
        result = repo.reconcile(thorough=True)
        with repo.lock_read():
            self.assertParentsMatch(scenario.corrected_parents(), repo,
                                    b'after')
            # The contents of the versions in the versionedfile should be the
            # same after the reconcile.
            self.assertEqual(
                vf_shas,
                self.shas_for_versions_of_file(
                    repo, scenario.all_versions_after_reconcile()))

            # Scenario.corrected_fulltexts contains texts which the test wants
            # to assert are now fulltexts. However this is an abstraction
            # violation; really we care that:
            # - the text is reconstructable
            # - it has an empty parents list
            # (we specify it this way because a store can use arbitrary
            # compression pointers in principle.
            for file_version in scenario.corrected_fulltexts():
                key = (b'a-file-id', file_version)
                self.assertEqual({key: ()}, repo.texts.get_parent_map([key]))
                self.assertIsInstance(
                    next(repo.texts.get_record_stream([key], 'unordered',
                                                      True)).get_bytes_as('fulltext'),
                    bytes)

    def test_check_behaviour(self):
        """Populate a repository and check it, and verify the output."""
        repo, scenario = self.prepare_test_repository()
        check_result = repo.check()
        check_result.report_results(verbose=True)
        log = self.get_log()
        for pattern in scenario.check_regexes(repo):
            self.assertContainsRe(log, pattern)

    def test_find_text_key_references(self):
        """Test that find_text_key_references finds erroneous references."""
        repo, scenario = self.prepare_test_repository()
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(scenario.repository_text_key_references(),
                         repo.find_text_key_references())

    def test__generate_text_key_index(self):
        """Test that the generated text key index has all entries."""
        repo, scenario = self.prepare_test_repository()
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(scenario.repository_text_key_index(),
                         repo._generate_text_key_index())



# Copyright (C) 2006, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A generator which creates a yaml stanza of the current tree info"""

import yaml

from breezy import errors, hooks
from breezy.revision import NULL_REVISION
from breezy.version_info_formats import VersionInfoBuilder, create_date_str


class YamlVersionInfoBuilder(VersionInfoBuilder):
    """This writes a yaml stream out."""

    def generate(self, to_file):
        info = {}
        revision_id = self._get_revision_id()
        if revision_id != NULL_REVISION:
            info['revision-id'] = revision_id.decode('utf-8')
            rev = self._branch.repository.get_revision(revision_id)
            info['date'] = create_date_str(rev.timestamp, rev.timezone)
            try:
                revno = self._get_revno_str(revision_id)
            except errors.GhostRevisionsHaveNoRevno:
                revno = None
            for hook in YamlVersionInfoBuilder.hooks['revision']:
                hook(rev, info)
        else:
            revno = '0'

        info['build-date'] = create_date_str()
        if revno is not None:
            info['revno'] = revno

        if self._branch.nick is not None:
            info['branch-nick'] = self._branch.nick

        if self._check or self._include_file_revs:
            self._extract_file_revisions()

        if self._check:
            if self._clean:
                info['clean'] = True
            else:
                info['clean'] = False

        if self._include_history:
            log = []
            for (revision_id, message,
                 timestamp, timezone) in self._iter_revision_history():
                log.append({
                    'id': revision_id.decode('utf-8'),
                    'message': message,
                    'date': create_date_str(timestamp, timezone),
                })
            info['revisions'] = log

        if self._include_file_revs:
            files = []
            for path in sorted(self._file_revisions.keys()):
                files.append({
                    'path': path,
                    'revision': (
                        self._file_revisions[path].decode('utf-8')
                        if isinstance(self._file_revisions[path], bytes) else self._file_revisions[path])
                })
            info['file-revisions'] = files

        yaml.dump(info, to_file)


class YamlVersionInfoBuilderHooks(hooks.Hooks):
    """Hooks for yaml-formatted version-info output."""

    def __init__(self):
        super().__init__(
            "breezy.version_info_formats.format_yaml", "YamlVersionInfoBuilder.hooks")
        self.add_hook('revision',
                      "Invoked when adding information about a revision to the"
                      " YAML stanza that is printed. revision is called with a"
                      " revision object and a YAML stanza.", (3, 3))


YamlVersionInfoBuilder.hooks = YamlVersionInfoBuilderHooks()  # type: ignore



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Front-end command for shell-like test scripts.

See doc/developers/testing.txt for more explanations.
This module should be importable even if testtools aren't available.
"""

import os

from . import commands, option


class cmd_test_script(commands.Command):
    """Run a shell-like test from a file."""

    hidden = True
    takes_args = ['infile']
    takes_options = [
        option.Option('null-output',
                      help='Null command outputs match any output.'),
        ]

    @commands.display_command
    def run(self, infile, null_output=False):
        # local imports to defer testtools dependency
        from breezy import tests
        from breezy.tests.script import TestCaseWithTransportAndScript

        with open(infile) as f:
            script = f.read()

        class Test(TestCaseWithTransportAndScript):

            script = None  # Set before running

            def test_it(self):
                self.run_script(script,
                                null_output_matches_anything=null_output)

        runner = tests.TextTestRunner(stream=self.outf)
        test = Test('test_it')
        test.path = os.path.realpath(infile)
        res = runner.run(test)
        return len(res.errors) + len(res.failures)



# Copyright (C) 2006, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Python implementation of readdir interface."""

import stat

_directory = 'directory'
_chardev = 'chardev'
_block = 'block'
_file = 'file'
_fifo = 'fifo'
_symlink = 'symlink'
_socket = 'socket'
_unknown = 'unknown'

_formats = {
    stat.S_IFDIR: 'directory',
    stat.S_IFCHR: 'chardev',
    stat.S_IFBLK: 'block',
    stat.S_IFREG: 'file',
    stat.S_IFIFO: 'fifo',
    stat.S_IFLNK: 'symlink',
    stat.S_IFSOCK: 'socket',
}


def _kind_from_mode(stat_mode, _formats=_formats, _unknown='unknown'):
    """Generate a file kind from a stat mode. This is used in walkdirs.

    It's performance is critical: Do not mutate without careful benchmarking.
    """
    try:
        return _formats[stat_mode & 0o170000]
    except KeyError:
        return _unknown
